@@ -4,7 +4,7 @@ use v5.14;
 use warnings;
 use utf8;
 
-our $VERSION = "2.29";
+our $VERSION = "2.30";
 
 use Data::Dumper;
 {
@@ -53,17 +53,16 @@ our $nonspace_re = qr{ \p{IsPrintableLatin} }x;
 our $reset_re    = qr{ \e \[ [0;]* m }x;
 our $color_re    = qr{ \e \[ [\d;]* m }x;
 our $erase_re    = qr{ \e \[ [\d;]* K }x;
-our $csi_body_re = qr{
-    # see ECMA-48 5.4 Control sequences
-    (?: \e\[ | \x9b )	# csi
-    [\x30-\x3f]*	# parameter bytes
-    [\x20-\x2f]*	# intermediate bytes
-}x;
-our $csi_re      = qr{
-    $csi_body_re
-    [\x40-\x7e]		# final byte
-}x;
-our $osc_re      = qr{
+
+# see ECMA-48 5.4 Control sequences
+my  $csi_start       = qr{ (?: \e\[ | \x9b ) }x;
+my  $csi_parameter   = qr/[\x30-\x3f]*+/;
+my  $csi_itermidiate = qr/[\x20-\x2f]*+/;
+my  $csi_final       = qr/[\x40-\x7e]/;
+our $csi_body_re     = qr/${csi_start}${csi_parameter}${csi_itermidiate}/;
+our $csi_re          = qr/${csi_start}${csi_parameter}${csi_itermidiate}${csi_final}/;
+
+our $osc_re          = qr{
     # see ECMA-48 8.3.89 OSC - OPERATING SYSTEM COMMAND
     (?: \e\] | \x9d )		# osc
     [\x08-\x13\x20-\x7d]*+	# command
@@ -85,6 +84,7 @@ sub IsWideSpacing {
 +utf8::East_Asian_Width=Wide
 +utf8::East_Asian_Width=FullWidth
 -utf8::Nonspacing_Mark
+-utf8::Default_Ignorable_Code_Point
 END
 }
 
@@ -94,6 +94,7 @@ sub IsWideAmbiguousSpacing {
 +utf8::East_Asian_Width=FullWidth
 +utf8::East_Asian_Width=Ambiguous
 -utf8::Nonspacing_Mark
+-utf8::Default_Ignorable_Code_Point
 END
 }
 
@@ -334,9 +335,15 @@ sub fold {
 	}
 	# carriage return
 	if (s/\A(\r+)//) {
-	    $folded .= $1;
-	    $room = $width;
-	    next;
+	    if (length == 0) {
+		# this must be the part of CRNL
+		$eol = $1;
+		last;
+	    } else {
+		$folded .= $1;
+		$room = $width;
+		next;
+	    }
 	}
 	# ECMA-48 OPERATING SYSTEM COMMAND
 	if (s/\A($osc_re)//) {
@@ -373,7 +380,7 @@ sub fold {
 	}
 
 	# imcomplete CSI
-	if (s/\A(\e|$csi_body_re)\z//) {
+	if (s/\A(\e+|$csi_body_re)\z//) {
 	    $folded .= $1;
 	    last;
 	}
@@ -609,7 +616,7 @@ Text::ANSI::Fold - Text folding library supporting ANSI terminal sequence and As
 
 =head1 VERSION
 
-Version 2.29
+Version 2.30
 
 =head1 SYNOPSIS
 
